/*
 * FareBotApplication.java
 *
 * Copyright (C) 2011 Eric Butler
 *
 * Authors:
 * Eric Butler <eric@codebutler.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.id.micolous.metrodroid;

import android.app.Application;
import android.nfc.NfcAdapter;
import android.os.StrictMode;
import android.preference.PreferenceManager;
import android.util.Log;

import com.codebutler.farebot.card.Card;
import com.codebutler.farebot.card.CardType;
import com.codebutler.farebot.card.classic.ClassicSector;
import com.codebutler.farebot.card.desfire.files.DesfireFile;
import com.codebutler.farebot.card.desfire.settings.DesfireFileSettings;
import com.codebutler.farebot.card.desfire.files.InvalidDesfireFile;
import com.codebutler.farebot.card.desfire.files.RecordDesfireFile;
import com.codebutler.farebot.card.felica.FelicaDBUtil;
import com.codebutler.farebot.transit.lax_tap.LaxTapDBUtil;
import com.codebutler.farebot.transit.ovc.OVChipDBUtil;
import com.codebutler.farebot.transit.seq_go.SeqGoDBUtil;
import com.codebutler.farebot.xml.Base64String;
import com.codebutler.farebot.xml.CardConverter;
import com.codebutler.farebot.xml.CardTypeTransform;
import com.codebutler.farebot.xml.ClassicSectorConverter;
import com.codebutler.farebot.xml.DesfireFileConverter;
import com.codebutler.farebot.xml.DesfireFileSettingsConverter;
import com.codebutler.farebot.xml.EpochDateTransform;
import com.codebutler.farebot.xml.FelicaIDmTransform;
import com.codebutler.farebot.xml.FelicaPMmTransform;
import com.codebutler.farebot.xml.HexString;
import com.codebutler.farebot.xml.SkippableRegistryStrategy;

import net.kazzz.felica.lib.FeliCaLib;

import org.simpleframework.xml.Serializer;
import org.simpleframework.xml.convert.Registry;
import org.simpleframework.xml.core.Persister;
import org.simpleframework.xml.strategy.Type;
import org.simpleframework.xml.strategy.Visitor;
import org.simpleframework.xml.strategy.VisitorStrategy;
import org.simpleframework.xml.stream.InputNode;
import org.simpleframework.xml.stream.NodeMap;
import org.simpleframework.xml.stream.OutputNode;
import org.simpleframework.xml.transform.RegistryMatcher;

import java.io.File;
import java.util.Date;

import au.id.micolous.farebot.R;

public class MetrodroidApplication extends Application {
    private static final String TAG = "MetrodroidApplication";
    public static final String PREF_LAST_READ_ID = "last_read_id";
    public static final String PREF_LAST_READ_AT = "last_read_at";
    public static final String PREF_MFC_AUTHRETRY = "pref_mfc_authretry";
    public static final String PREF_MFC_FALLBACK = "pref_mfc_fallback";

    private static MetrodroidApplication sInstance;

    private FelicaDBUtil mFelicaDBUtil;
    private OVChipDBUtil mOVChipDBUtil;
    private SeqGoDBUtil mSeqGoDBUtil;
    private LaxTapDBUtil mLaxTapDBUtil;
    private final Serializer mSerializer;
    private boolean mHasNfcHardware = false;
    private boolean mMifareClassicSupport = false;

    public MetrodroidApplication() {
        sInstance = this;

        mFelicaDBUtil = new FelicaDBUtil(this);
        mOVChipDBUtil = new OVChipDBUtil(this);
        mSeqGoDBUtil = new SeqGoDBUtil(this);
        mLaxTapDBUtil = new LaxTapDBUtil(this);

        try {
            Visitor visitor = new Visitor() {
                @Override public void read(Type type, NodeMap<InputNode> node) throws Exception { }
                @Override public void write(Type type, NodeMap<OutputNode> node) throws Exception {
                    node.remove("class");
                }
            };
            Registry registry = new Registry();
            RegistryMatcher matcher = new RegistryMatcher();
            mSerializer = new Persister(new VisitorStrategy(visitor, new SkippableRegistryStrategy(registry)), matcher);

            DesfireFileConverter desfireFileConverter = new DesfireFileConverter(mSerializer);
            registry.bind(DesfireFile.class, desfireFileConverter);
            registry.bind(RecordDesfireFile.class, desfireFileConverter);
            registry.bind(InvalidDesfireFile.class, desfireFileConverter);

            registry.bind(DesfireFileSettings.class, new DesfireFileSettingsConverter());
            registry.bind(ClassicSector.class, new ClassicSectorConverter());
            registry.bind(Card.class, new CardConverter(mSerializer));

            matcher.bind(HexString.class, HexString.Transform.class);
            matcher.bind(Base64String.class, Base64String.Transform.class);
            matcher.bind(Date.class, EpochDateTransform.class);
            matcher.bind(FeliCaLib.IDm.class, FelicaIDmTransform.class);
            matcher.bind(FeliCaLib.PMm.class, FelicaPMmTransform.class);
            matcher.bind(CardType.class, CardTypeTransform.class);

        } catch (Exception ex) {
            throw new RuntimeException(ex);
        }
    }

    public static MetrodroidApplication getInstance() {
        return sInstance;
    }

    public FelicaDBUtil getFelicaDBUtil() {
        return mFelicaDBUtil;
    }

    public OVChipDBUtil getOVChipDBUtil() {
        return mOVChipDBUtil;
    }

    public SeqGoDBUtil getSeqGoDBUtil() {
        return mSeqGoDBUtil;
    }

    public LaxTapDBUtil getLaxTapDBUtil() {
        return mLaxTapDBUtil;
    }

    public Serializer getSerializer() {
        return mSerializer;
    }

    public boolean getMifareClassicSupport() { return mMifareClassicSupport; }

    @Override public void onCreate() {
        super.onCreate();

        try {
            detectNfcSupport();
        } catch (Exception e) {
            Log.w(TAG, "Detecting nfc support failed", e);
        }

        PreferenceManager.setDefaultValues(this, R.xml.prefs, false);

        StrictMode.setThreadPolicy(new StrictMode.ThreadPolicy.Builder()
            .detectAll()
            .penaltyLog()
            .build());

    }

    private void detectNfcSupport() {
        // Some devices like the LG F60 misreport they support Mifare Classic when they don't.
        // Others report they don't support Mifare Classic when they do.

        // TODO: determine behaviour of Microread hardware. It may support MFC.
        File device;
        NfcAdapter nfcAdapter = NfcAdapter.getDefaultAdapter(this);
        mHasNfcHardware = nfcAdapter != null;

        if (!mHasNfcHardware) {
            return;
        }

        // Check for Broadcom NFC controller
        // == no MFC support
        device = new File("/dev/bcm2079-i2c");
        if (device.exists()) {
            Log.d(TAG, "Detected Broadcom bcm2079");
            mMifareClassicSupport = false;
            return;
        }

        // Check for NXP pn544 NFC controller
        // == has MFC support
        device = new File("/dev/pn544");
        if (device.exists()) {
            Log.d(TAG, "Detected NXP pn544");
            mMifareClassicSupport = true;
            return;
        }

        // Check for shared libraries corresponding to non-NXP chips.
        File libFolder = new File("/system/lib");
        File[] libs = libFolder.listFiles();
        for (File lib : libs) {
            String name = lib.getName();
            if (lib.isFile() && name.startsWith("libnfc") && name.contains("brcm")) {
                Log.d(TAG, "Detected Broadcom NFC library");
                mMifareClassicSupport = false;
                return;
            }
        }

        // Fallback: Look for com.nxp.mifare feature.
        mMifareClassicSupport = this.getPackageManager().hasSystemFeature("com.nxp.mifare");
        Log.d(TAG, "Falling back to com.nxp.mifare feature detection " +
                (mMifareClassicSupport ? "(found)" : "(missing)"));
    }

}
